// max values by output file
#define OUTPUT_MAX_VAL 1000
#define OUTPUT_CSV_SEP ";"

#define MAXFPS 40
#define MAXCPS 500

#define DEFAULT_PLANETS_FILE "planets.txt"

struct date
{
    int centuries5;
    double timestamp;

    date() : centuries5(0), timestamp(0.0) {}
};

extern date simul;

struct OrbitDatas
{
    double aphelion, perihelion, period;
    bool hasperiod;
    vec aph, per;
    OrbitDatas() : aphelion(0), perihelion(0), period(0), hasperiod(false) { loop(i, 3) aph.v[i] = per.v[i] = 0; }
};

enum { BODY_PL = 0, BODY_SAT, BODY_NUM };

class Body
{
     public: 
        string name;
        vec pos, v, a, push;
        double mass; // kg
        double radius, histupdate;
        short int r, g, b;
        int type;
    
     Concurrency::concurrent_vector<vec *> hist;
};

class Planet : public Body
{
    public:
        vec refpos, refposcenter; // position, first position, velocity, acceleration
        bool histshift, showhist;
        double approaching, curtime;
        vec curangle, curref;
        date refposdate;
        double angle, area;

        svector<OrbitDatas> orbits;

    void resethist();

    Planet();
    ~Planet();
};

enum { SAT_ACCEL = 0, SAT_APPROACH, SAT_ORBIT };

class Satellite : public Body
{
    public: 
        vec push;
        vec goal_p, goal_v; // ideal position and velocity to reach
        vec axis; // axis the satellite is rotating around.
        double goal_vel, max_vel; // orbit velocity, max velocity
        double period, free_flight, stop_time;
        short int state;
        
        double max_push, max_brake; // newtons
        bool orbiting; // has reached orbiting position ?
        Planet *pl; // planet the satellite is orbiting around;

    void resethist();

    Satellite();
    ~Satellite();
};

struct threadmillis
{
    int last, current, diff, fpserror, maxfps;
    threadmillis() : last(0), current(0), diff(0), fpserror(0), maxfps(100) {}
};

extern threadmillis render_millis, compute_millis;

bool load_planets(const char *filename);
void save_planets();
void save_datas();
void clear_hist();
void move_planets(int it);
double calc_angle(Planet *pl);
double calc_area(Planet *pl);

extern threadmillis render_millis, compute_millis;
extern int lasthistclear;
extern int reference;
extern double scalefactor;
extern bool running, print_frame;
extern char *planets_file;
extern Planet *curplanet;
extern svector<Planet *> planets;
extern svector<Satellite *> satellites;
extern double precision, speed, seconds_per_dot;
extern bool skip_hists, draw_accel, draw_area;

struct unit
{
    double d_factor, t_factor, m_factor; // distance factor, time factor, masse factor (with MKS)
    string d_unit, t_unit, m_unit;
};

enum { UNITS_MKS = 0, UNITS_UAKJ, UNITS_NUM };

extern unit units[UNITS_NUM];
extern int units_system;


// i/o controls

struct Zone
{
    int x, y, xs, ys;
    
    Zone() : x(0), y(0), xs(0), ys(0) {}
    Zone(int x, int y, int xs, int ys) : x(x), y(y), xs(xs), ys(ys) {}
};

#define BUTTON_DEFAULT_WIDTH 32
#define BUTTON_DEFAULT_HEIGHT 32

enum { B_STARTPAUSE = 0, B_FASTER, B_SLOWER,
    B_ZOOMIN, B_ZOOMOUT,
    B_PREVREF, B_NEXTREF,
    B_MOREDOTS, B_LESSDOTS,
    B_STUDYNEXT,
    B_REWIND,
    B_NUM };

struct Button
{
    string name;
    string label;
    bool show, available;
    int tex_x, tex_y, width, height, x, y;
    void (*action)(void);
    bool pressed, hover;
};

void init_buttons();
void start_pause(int force);
void start_pause();
void increase_speed();
void lower_speed();
void zoom_in();
void zoom_out();
void prev_ref();
void next_ref();
void smoother();
void harder();
void studied();
void rewind();

extern Button buttons[B_NUM];
extern GLuint buttons_tex;
extern int buttons_width;

extern int mouse_x, mouse_y;
void mouse_click(short int mouse_button, int mx, int my, bool isdown);